import multiprocessing
import threading
import queue
import gc
from typing import Type, Dict, Tuple, Any, NoReturn
from concurrent.futures import Future
from uuid import uuid4, UUID
from typing import NamedTuple, Type

import numpy as np

from vcap import BaseBackend, DETECTION_NODE_TYPE


class RpcRequest(NamedTuple):
    request_id: str
    """A string UUID"""
    function: str
    """The function to run"""
    args: Tuple[Any]
    """Arguments for the function"""
    kwargs: Dict[str, Any]
    """Keyword arguments for the function"""


class RpcResponse(NamedTuple):
    request_id: str
    """A string UUID corresponding to the RpcRequest"""
    result: Any
    """The result of the function that was run"""
    exception: BaseException
    """If the function threw an exception, this will hold the exception. 
    In the case when the function failed with an exceptio, RpcResponse.result
    will be None."""


def _rpc_server(
        incoming: multiprocessing.Queue,
        outgoing: multiprocessing.Queue,
        shutdown: multiprocessing.Event,
        num_workers: int,
        backend_class: Type[BaseBackend],
        args,
        kwargs):
    """ An RPC server for running Backend functions.
    :param incoming: Receives RpcRequests through here
    :param outgoing: Returns RpcResponses here
    :param shutdown: When set, this signals that the server should shut down.
    :param num_workers: How many internal threads should exist for processing
    incoming RpcRequests.
    :param backend_class: The backend to initialize
    :param args: *args for the backend_class __init__
    :param kwargs: *kwargs for the backend_class __init__
    """
    # Initialize the backend and catch any errors during initialization
    try:
        backend = backend_class(*args, **kwargs)
    except BaseException as e:
        outgoing.put(e)
        return
    else:
        outgoing.put(None)

    pool = multiprocessing.pool.ThreadPool(processes=num_workers)

    def handle_request(request: RpcRequest):
        try:
            fn = getattr(backend, request.function)  # noqa: F821
            result = fn(*request.args, **request.kwargs)
        except BaseException as e:
            result = None
            exception = e
        else:
            exception = None

        outgoing.put(RpcResponse(
            request_id=request.request_id,
            result=result,
            exception=exception
        ))

    while not shutdown.is_set() or incoming.qsize():
        try:
            request: RpcRequest = incoming.get(timeout=0.1)
        except queue.Empty:
            continue
        pool.apply_async(handle_request, args=(request,))
    pool.close()
    pool.join()
    del backend
    # Make sure any __del__'s gets called, potentially freeing up
    # sockets that the backend may have been connected to. For example,
    # the OpenVINO HDDL plugin connects to a socket, and releases it
    # upon garbage collection.
    gc.collect()


class BackendRpcProcess(BaseBackend):
    """Runs a backend inside of another process.
    This functionality is very experimental! It may not work for all cases
    or frameworks.

    NOTE! This backend wrapper currently can only be used on capsules
    that do not modify detection nodes in-place. Most if not all detectors
    should work fine.
        This can potentially be fixed in the future by diffing the
    node.__dict__ of the incoming vs outgoing node and making the appropriate
    additions.
    """

    def __init__(self, backend_class: Type[BaseBackend], *args, **kwargs):
        """
        :param backend_class: The Backend type to initialize
        :param args: Arguments for the backend
        :param kwargs: Keyword arguments for the backend
        """
        self._backend_class = backend_class
        self._incoming = multiprocessing.Queue()
        self._outgoing = multiprocessing.Queue()
        self._shutdown = multiprocessing.Event()
        self._futures: Dict[str, Future] = {}
        """Keep track of ongoing requests in a dict of request_id: Future """

        # Spin up the server process
        self._process = multiprocessing.Process(
            target=_rpc_server,
            daemon=True,
            name="RpcServer",
            kwargs={
                "outgoing": self._incoming,
                "incoming": self._outgoing,
                "shutdown": self._shutdown,
                "backend_class": backend_class,
                "num_workers": 100,
                "args": args,
                "kwargs": kwargs
            })
        self._process.start()
        # Wait for success or failure of the Backend initialization
        exception = self._incoming.get()
        if exception:
            raise exception

        # Start the client thread
        self._rpc_thread = threading.Thread(
            target=self._rpc_client,
            name="RpcClientThread")
        self._rpc_thread.start()

    @property
    def workload(self) -> int:
        # TODO: Use RPC to request the underlying Backends
        #       'workload' implementation
        return len(self._futures)

    def _rpc_client(self):
        """This is the dedicated thread for receiving and routing results
        from the remote backend. It receives incoming RpcResponses and
        routes results or exceptions to the relevant Future objects."""
        while not self._shutdown.is_set() or self._incoming.qsize():
            try:
                response: RpcResponse = self._incoming.get(timeout=0.1)
            except queue.Empty:
                continue
            future = self._futures.pop(response.request_id)
            if response.exception:
                future.set_exception(response.exception)
            else:
                future.set_result(response.result)

    def _rpc_call(self, function: str, *args, **kwargs) -> Any:
        """Run a function on the remote backend.
        :param function: The function to run on the remote backend
        :param args: Arguments for the function
        :param kwargs: Keyword arguments for the function
        """
        request = RpcRequest(
            function=function,
            args=args,
            kwargs=kwargs,
            request_id=uuid4().hex
        )

        future = Future()
        self._futures[request.request_id] = future
        self._outgoing.put(request)
        return future.result()

    def process_frame(self, *args, **kwargs) -> DETECTION_NODE_TYPE:
        return self._rpc_call("process_frame", *args, **kwargs)

    def distances(self, *args, **kwargs) -> np.ndarray:
        return self._rpc_call("distances", *args, **kwargs)

    def batch_predict(self, *args, **kwargs) -> NoReturn:
        """This function is implemented on the backend running on the backend
        process. It shouldn't be called from the parent process."""
        raise NotImplementedError()

    def close(self) -> None:
        # Close the underlying backend
        self._rpc_call("close")
        # Close the server process and the client thread
        self._shutdown.set()
        self._process.join()
        self._rpc_thread.join()



from .iou_cost_matrix import detection_area, detection_intersection, \
    detection_iou, bbox_intersection, bbox_iou, iou_cost_matrix
from .linear_assignment import linear_assignment
from .nonmax_suppression import non_max_suppression
from .distances import cosine_distance, euclidian_distance



import logging
from functools import partial
from itertools import cycle
from threading import Event
from typing import Dict, List, Tuple, Deque
from collections import deque

import numpy as np

from vcap import Resize, BaseBackend, DetectionNode

_SUPPORTED_METRICS = "SUPPORTED_METRICS"
_RANGE_FOR_ASYNC_INFER_REQUESTS = "RANGE_FOR_ASYNC_INFER_REQUESTS"
OV_INPUT_TYPE = Dict[str, np.ndarray]


class BaseOpenVINOBackend(BaseBackend):
    def __init__(self, model_xml: bytes,
                 weights_bin: bytes,
                 device_name: str,
                 ie_core=None):
        """
        :param model_xml: The XML data defining the OpenVINO model architecture
        :param weights_bin: The .bin file data defining the model's weights
        :param ie_core: :
          None (default): The backend will initialize its own IECore to load
          the network with.
          IECore: An initialized openvino.inference_engine.IECore, with any
          settings already applied. This can be used to apply CPU extensions
          or load different plugins to the IECore giving it to the backend.
        """
        super().__init__()
        # Convert from the vcap device naming format to openvino format
        device_name = "CPU" if device_name[:4] == "CPU:" else device_name

        from openvino.inference_engine import \
            IECore, ExecutableNetwork, IENetwork, StatusCode

        self.ie = ie_core or IECore()

        is_multi_device = device_name.lower().startswith("multi")

        # Find the optimal number of InferRequests for this device
        n_requests = 0
        if not is_multi_device:
            supported_metrics = self.ie.get_metric(
                device_name, _SUPPORTED_METRICS)
            if _RANGE_FOR_ASYNC_INFER_REQUESTS in supported_metrics:
                low, high, _ = self.ie.get_metric(
                    device_name, _RANGE_FOR_ASYNC_INFER_REQUESTS)
                # Cap the n_requests, because setting it too high can crash
                # TODO(Alex): Figure out _why_ hddl crashes when set to 'high'
                n_requests = max(0, min(low * 2, high))

        self.net: IENetwork = self.ie.read_network(
            model=model_xml,
            weights=weights_bin,
            init_from_buffer=True)

        try:
            self.exec_net: ExecutableNetwork = self.ie.load_network(
                network=self.net,
                device_name=device_name,
                num_requests=n_requests)
        except RuntimeError as e:
            # This error happens when trying to load onto HDDL or MYRIAD, but
            # somehow the device fails to work. In these cases, we try again
            # but load onto 'CPU', which is a safe choice.

            if device_name == "CPU":
                # It's unknown why this would happen when loading
                # onto CPU, so we re-raise the error.
                raise e
            msg = f"Failed to load {self.__class__} onto device " \
                  f"{device_name}. Error: '{e}'. " \
                  f"Trying again with device_name='CPU'"
            logging.critical(msg)
            self.__init__(
                model_xml=model_xml,
                weights_bin=weights_bin,
                device_name="CPU",
                ie_core=ie_core)
            return

        # Pull out a couple useful constants
        self.InferRequestStatusCode = StatusCode
        self.input_blob_names: List[str] = list(self.net.inputs.keys())
        self.output_blob_names: List[str] = list(self.net.outputs.keys())

    def prepare_inputs(self, frame: np.ndarray, frame_input_name: str = None) \
            -> Tuple[OV_INPUT_TYPE, Resize]:
        """A helper method to create an OpenVINO input like {input_name: array}

        This method takes a frame, resizes it to fit the network inputs, then
        returns two things: The input, and the Resize information. The
        Resize information contains all of the operations that were done on
        the frame, allowing users to then map the detections from a resized
        frame to the coordinate space of the original frame.

        :param frame: The image. BGR ordered.
        :param frame_input_name: Set this value to force a certain node to be
            used as the frame input. Useful if you still want to use the
            default implementation from a subclass with network with multiple
            inputs
        :returns: ({input_name: resized_frame}, Resize)
        """

        if not frame_input_name and len(self.net.inputs) > 1:
            raise ValueError("More than one input was expected for model, but "
                             "default prepare_inputs implementation was used.")

        input_blob_name = frame_input_name or self.input_blob_names[0]
        input_blob = self.net.inputs[input_blob_name]

        _, _, h, w = input_blob.shape
        resize = Resize(frame).resize(w, h, Resize.ResizeType.EXACT)

        # Change data layout from HWC to CHW
        in_frame = np.transpose(resize.frame.copy(), (2, 0, 1))

        return {input_blob_name: in_frame}, resize

    def parse_detection_results(
            self, results: np.ndarray,
            resize: Resize,
            label_map: Dict[int, str],
            min_confidence: float = 0.0) -> List[DetectionNode]:
        """A helper method to take results from a detection-type network.
        :param results: The inference results from the network
        :param resize: A Resize object that was used to resize the image to
        fit into the network originally.
        :param label_map: A dictionary mapping integers to class_names.
        :param min_confidence: Filter out detections that have a confidence
        less than this number.
        :returns: A list of DetectionNodes, in this case representing bounding
        boxes.
        """
        output_blob_name = self.output_blob_names[0]
        inference_results = results[output_blob_name]

        _, _, h, w = self.net.inputs[self.input_blob_names[0]].shape

        nodes: List[DetectionNode] = []
        for result in inference_results[0][0]:
            # If the first index == 0, that's the end of real predictions
            # The network always outputs an array of length 200 even if it does
            # not have that many predictions
            if result[0] != 0:
                break

            confidence = float(result[2])
            if confidence <= min_confidence:
                continue

            x_min, y_min, x_max, y_max = result[3:7]
            # x and y in res are in terms of percent of image width/height
            x_min, x_max = x_min * w, x_max * w
            y_min, y_max = y_min * h, y_max * h
            coords = [[x_min, y_min], [x_max, y_min],
                      [x_max, y_max], [x_min, y_max]]

            class_id = round(result[1])
            res = DetectionNode(
                name=label_map[class_id],
                coords=coords,
                extra_data={"detection_confidence": confidence})
            nodes.append(res)

        # Convert the coordinate space of the detections from the
        # resized frame to the
        resize.scale_and_offset_detection_nodes(nodes)
        return nodes

    def batch_predict(self, inputs: List[OV_INPUT_TYPE]) \
            -> List[object]:
        """Use the network for inference.

        This function will receive a list of inputs and process them as
        efficiently as possible, optimizing for throughput.
        :param inputs: A list of openvino style inputs {input_name: ndarray}
        :returns: A generator of the networks outputs, yielding them in the
        same order as the inputs
        """
        n_inputs: int = len(inputs)
        """Keep track of the number of expected total results"""
        next_frame_id: int = 0
        """The next frame_id we are awaiting results to send. This guarantees
        that results are sent in the same order as the inputs."""
        inputs: Deque[Tuple[int, OV_INPUT_TYPE]] = deque(enumerate(inputs))
        """A deque containing tuples of (frame_id, input) for inference"""
        unsent_results: Dict[int: Dict] = {}
        """A dictionary of {frame_id: output}, the results not yet yielded"""
        requests = [(Event(), request) for request in self.exec_net.requests]
        """A list of (Event, InferRequest) tuples. The Event is set when 
         the InferRequest is finished with inference, and at the start of
         the run."""

        def on_result(request_free_event, request, frame_id, *args):
            unsent_results[frame_id] = request.outputs
            request_free_event.set()

        # Start all request_events as set (ie, ready for inference)
        for request_free, _ in requests:
            request_free.set()

        # This loop will end when all inputs have been processed and outputs
        # have been yielded
        for request_free, request in cycle(requests):
            # Return any results that are waiting to be returned
            while next_frame_id in unsent_results:
                yield unsent_results.pop(next_frame_id)
                next_frame_id += 1

            if next_frame_id == n_inputs:
                break

            # Wait for the request to be done with previous inference
            request_free.wait()

            if not len(inputs):
                continue

            # Put another request in the queue, if there are frames
            frame_id, input_dict = inputs.popleft()
            request.set_completion_callback(
                partial(on_result, request_free, request, frame_id))
            request_free.clear()
            request.async_infer(input_dict)

    def close(self):
        super().close()
        # Since there's no way to tell OpenVINO to close sockets to HDDL
        # (or other plugins), dereferencing everything is the safest way
        # to go. Without this, OpenVINO seems to crash the HDDL daemon.
        self.ie = None
        self.net = None
        self.exec_net = None



import cv2
import numpy as np
import tensorflow as tf

from .base_tensorflow import BaseTFBackend
from .predictions import DepthPrediction
from .load_utils import parse_tf_model_bytes


class DepthPredictor(BaseTFBackend):
    """
    Loads a model and uses it to run depth prediction, meaning that it takes
    an image as input and returns a matrix the size of the input image where
    the value of each 'pixel' corresponds to the pixel's distance from the
    camera in meters.

    Does not support batch prediction TODO: Is this true?
    """

    def __init__(self, model_bytes,
                 device: str=None,
                 session_config: tf.ConfigProto=None):
        """
        :param model_bytes: Model file data, likely a loaded *.pb file
        :param device: The device to run the model on
        :param session_config: Model configuration options
        """
        super().__init__()
        self.graph, self.session = parse_tf_model_bytes(model_bytes,
                                                        device,
                                                        session_config)

        # Pull all the necessary inputs from the network
        self.input_image = self.graph.get_tensor_by_name("input:0")

        # Pull the necessary attributes that we want to get from the network
        # after running it
        self.segmented_tensor = self.graph.get_tensor_by_name(
            "output_prediction:0")

    def batch_predict(self, img_bgr) -> DepthPrediction:
        """
        Takes numpy BGR images of the format that OpenCV gives, and returns
        predicted labels in the form of a list of DetectionPrediction
        objects from predictions.py

        :param img_bgr: A numpy BGR image from OpenCV
        :return: List of list of DetectionPrediction objects
        """
        img_rgb = cv2.cvtColor(img_bgr, cv2.COLOR_BGR2RGB)
        feed = {self.input_image: img_rgb}
        segmented_image = self.session.run(self.segmented_tensor,
                                           feed_dict=feed)

        out = np.squeeze(segmented_image)

        return DepthPrediction(out)
